using UnityEngine;
using System.Collections;
using System.Linq;

[RequireComponent(typeof(AudioSource))]
public class MusicPlayer : MonoBehaviour
{
    [SerializeField] private AudioClip menuSong;
    [SerializeField] private AudioClip[] ingameSongs;
    [SerializeField] private AudioClip playerWonSong;
    [SerializeField] private AudioClip playerLostSong;
    [SerializeField] private float fadingToMenu = 2f;
    [SerializeField] private float fadingToIngame = 1f;
    [SerializeField] private float fadingToPlayerWon = 0.5f;
    [SerializeField] private float fadingToPlayerLost = 0.5f;

    private int ingameSongIndex;
    private AudioSource audioSource;

    private bool gameStarted;

    private AudioClip nextSong;
    private float currentFadingOutTime;
    private float fadingOutTimeLeft;
    private float _volume;

    void Awake()
    {
        audioSource = GetComponent<AudioSource>();

        ingameSongIndex = 0;
        gameStarted = false;

        Volume = 1;

        PlayMenuSong();
    }

    void Update()
    {
        // Fade the current song out and switch to the next song
        if (nextSong != null)
        {
            fadingOutTimeLeft -= Time.deltaTime;
            float volumePercent = FadingVolumePercent;

            if (volumePercent == 0)
            {
                audioSource.volume = Volume;
                audioSource.clip = nextSong;
                audioSource.Play();

                nextSong = null;
            }
            else
            {
                audioSource.volume = Volume * volumePercent;
            }

            return;
        }

        // If we are ingame and there is no song playing, switch to the next ingame song
        if ((!audioSource.isPlaying) && gameStarted)
        {
            ingameSongIndex = (ingameSongIndex + 1) % ingameSongs.Length;
            audioSource.clip = ingameSongs[ingameSongIndex];
            audioSource.Play();
        }
    }

    public void NewGameWasInitialized()
    {
        if (!gameStarted)
            return;

        gameStarted = false;

        PlayMenuSong();
    }

    private void PlayMenuSong()
    {
        audioSource.loop = true;
        SwitchToSong(menuSong, fadingToMenu);
    }

    public void SwitchedToIngame()
    {
        if (gameStarted)
            return;

        gameStarted = true;

        ingameSongIndex = 0;

        audioSource.loop = false;
        SwitchToSong(ingameSongs[ingameSongIndex], fadingToIngame);
    }

    public void PlayerWon()
    {
        SwitchToSong(playerWonSong, fadingToPlayerWon);
    }

    public void PlayerLost()
    {
        SwitchToSong(playerLostSong, fadingToPlayerLost);
    }

    public void SwitchToSong(AudioClip clip, float fadingOutTime)
    {
        if ((audioSource.isPlaying) && (fadingOutTime > 0))
        {
            nextSong = clip;
            fadingOutTimeLeft = fadingOutTime;
            currentFadingOutTime = fadingOutTime;
        }
        else
        {
            audioSource.clip = clip;
            audioSource.Play();
            nextSong = null;
            currentFadingOutTime = 0;
        }
    }

    public float Volume
    {
        get { return _volume; }
        set
        {
            _volume = value;
            audioSource.volume = FadingVolumePercent * _volume;
        }
    }

    private float FadingVolumePercent { get { return (nextSong == null) ? 1 : Mathf.Clamp01(fadingOutTimeLeft / currentFadingOutTime); } }

    public static MusicPlayer FindInstance()
    {
        return UnityHelper.FindFirstActiveInstance<MusicPlayer>();
    }
}
