﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using UnityEngine;
using Random = UnityEngine.Random;

public static class MathHelper
{
    public const double EPSILON = 0.0001f;

    public static Vector2 GetEllipsePointFromAngle(float angle, float a, float b)
    {
        // See: http://stackoverflow.com/questions/11309596/how-to-get-a-point-on-an-ellipses-outline-given-an-angle

        var c = Mathf.Cos(angle);
        var s = Mathf.Sin(angle);
        var ta = s / c;
        var tt = ta * a / b;
        var d = 1.0f / Mathf.Sqrt(1f + tt * tt);
        var x = CopySign(a * d, c);
        var y = CopySign(b * tt * d, s);

        return new Vector2(x, y);
    }

    // Return copySignTo with the sign of copySignFrom
    public static float CopySign(float copySignTo, float copySignFrom)
    {
        return Mathf.Abs(copySignTo) * Mathf.Sign(copySignFrom);
    }

    public static bool IsPointBoundedByEllipse(float pointX, float pointY, float ellipseX, float ellipseY, float a, float b)
    {
        // See: http://math.stackexchange.com/questions/76457/check-if-a-point-is-within-an-ellipse

        float dx = pointX - ellipseX;
        float dy = pointY - ellipseY;
        float v = (dx * dx) / (a * a) + (dy * dy) / (b * b);

        return v <= 1;
    }

    public static Vector3 ClosestPointOnLine(Vector3 vA, Vector3 vB, Vector3 vPoint)
    {
        // See: http://forum.unity3d.com/threads/8114-math-problem?p=59715&viewfull=1#post59715

        var vVector1 = vPoint - vA;
        var vVector2 = (vB - vA).normalized;
     
        var d = Vector3.Distance(vA, vB);
        var t = Vector3.Dot(vVector2, vVector1);
     
        if (t <= 0)
            return vA;
     
        if (t >= d)
            return vB;
     
        var vVector3 = vVector2 * t;
     
        var vClosestPoint = vA + vVector3;

        return vClosestPoint;
    }

    public static float[] GetAnglesOfProjectileTrajectory(Vector2 targetPosition, float speed)
    {
        // See: http://stackoverflow.com/questions/1972315/need-help-deciphering-a-formula-for-projectile-motion

        var x = targetPosition.x;
        var y = targetPosition.y;
        var gravity = -Physics.gravity.y;

        var tmp = Mathf.Pow(speed, 4) - gravity * (gravity * Mathf.Pow(x, 2) + 2 * y * Mathf.Pow(speed, 2));

        if (tmp < 0)
            return null;

        var angle1 = Mathf.Atan2(Mathf.Pow(speed, 2) + Mathf.Sqrt(tmp), gravity * x);
        var angle2 = Mathf.Atan2(Mathf.Pow(speed, 2) - Mathf.Sqrt(tmp), gravity * x);

        return new[] { angle1, angle2 };
    }

    public static bool IsEqualOrBetween(float value, float min, float max)
    {
        return (min <= value) && (value <= max);
    }

    public static float ClampOrCenterBetweenRanges(float value, float min, float max)
    {
        if (min >= max)
        {
            return (min + max) / 2f;
        }

        if (value < min)
        {
            return min;
        }

        if (value > max)
        {
            return max;
        }

        return value;
    }

    public static float LerpOverflow(float a, float b, float percent)
    {
        float delta = b - a;
        return a + delta * percent;
    }

    public static float GetCenterAngleDeg(float angle1, float angle2)
    {
        angle1 = NormalizeAngleDeg360(angle1);
        angle2 = NormalizeAngleDeg360(angle2);

        float center = (angle1 + angle2) / 2f;

        // If the delta is too big, take the other (nearer) center
        if (Mathf.Abs(angle1 - angle2) > 180)
            center -= 180;

        return center;
    }

    public static float NormalizeAngleDeg360(float angle)
    {
        while (angle < 0)
            angle += 360;

        if (angle > 360)
            angle %= 360;

        return angle;
    }

    public static float NormalizeAngleDeg180(float angle)
    {
        while (angle < -180)
            angle += 360;

        while (angle > 180)
            angle -= 360;

        return angle;
    }

    public static float Map(float sourceValue, float sourceFrom, float sourceTo, float targetFrom, float targetTo)
    {
        float sourceRange = sourceTo - sourceFrom;
        float targetRange = targetTo - targetFrom;

        float percent = Mathf.Clamp01((sourceValue - sourceFrom) / sourceRange);

        float targetValue = targetFrom + targetRange * percent;

        return targetValue;
    }

    public static float MapJoystick(float sourceValue, float sourceFrom, float sourceTo, float deadzone = 0f, bool fullRangeBetweenDeadzoneAndOne = false)
    {
        var percent = Map(sourceValue, sourceFrom, sourceTo, -1, 1);

        if (Mathf.Abs(percent) <= deadzone)
            return 0;

        if (fullRangeBetweenDeadzoneAndOne && (deadzone > 0f))
        {
            if (percent < 0)
            {
                return Map(percent, -1f, -deadzone, -1f, 0f);
            }
            else
            {
                return Map(percent, deadzone, 1f, 0f, 1f);
            }
        }

        return percent;
    }

    public static Vector2 RandomOnUnitCircle
    {
        get
        {
            Vector2 result;
            do
            {
                result = Random.insideUnitCircle;
            } while (result.sqrMagnitude == 0);

            return result.normalized;
        }
    }
}
