﻿using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using Object = UnityEngine.Object;

public static class UnityHelper
{
    public static void CopyPositionAndRotatationFrom(this Transform transform, Transform source)
    {
        transform.position = source.position;
        transform.rotation = source.rotation;
    }

    public static T FindFirstActiveInstance<T>() where T: Component
    {
        var instances = (T[]) UnityEngine.Object.FindObjectsOfType(typeof (T));
        return instances.FirstOrDefault(instance => instance.gameObject.activeInHierarchy);
    }

    public static void SetTextColor(this TextMesh textMesh, Color color)
    {
        var textRenderer = textMesh.GetComponent<MeshRenderer>();
        textRenderer.material = new Material(textRenderer.material);
        textRenderer.material.color = color;
    }

    public static T GetComponentByInterface<T>(this Component component) where T : class
    {
        return component.GetComponents<Component>().OfType<T>().FirstOrDefault();
    }

    public static T[] GetComponentsByInterface<T>(this Component component) where T : class
    {
        return component.GetComponents<Component>().OfType<T>().ToArray();
    }

    public static T GetComponentParentHierachy<T>(this Component startPoint) where T : Component
    {
        while (startPoint != null)
        {
            var component = startPoint.GetComponent<T>();
            if (component != null)
                return component;

            startPoint = startPoint.transform.parent;
        }

        return null;
    }

    public static void AssignLayerToHierarchy(this GameObject gameObject, int layer)
    {
        foreach (var transform in gameObject.GetComponentsInChildren<Transform>())
        {
            transform.gameObject.layer = layer;
        }
    }

    public static Vector2 CalculateViewportSizeAtDistance(this Camera camera, float distance, float aspectRatio = 0)
    {
        if (aspectRatio == 0)
        {
            aspectRatio = camera.aspect;
        }

        var viewportHeightAtDistance = 2.0f * Mathf.Tan(0.5f * camera.fieldOfView * Mathf.Deg2Rad) * distance;
        var viewportWidthAtDistance = viewportHeightAtDistance * aspectRatio;

        return new Vector2(viewportWidthAtDistance, viewportHeightAtDistance);
    }

    public static bool ContainsLayer(this LayerMask mask, int layer)
    {
        return (mask.value & (1 << layer)) != 0;
    }

    public static GameObject[] FindGameObjectsWithLayer(int layer)
    {
        var goArray = (GameObject[]) UnityEngine.Object.FindObjectsOfType(typeof (GameObject));
        var results = goArray.Where(t => t.layer == layer).ToList();
        return results.ToArray();
    }

    public static GameObject[] FindGameObjectsWithLayer(LayerMask layerMask)
    {
        var goArray = (GameObject[]) UnityEngine.Object.FindObjectsOfType(typeof (GameObject));
        var results = goArray.Where(t => layerMask.ContainsLayer(t.layer)).ToList();
        return results.ToArray();
    }

    public static Bounds CreateBoundsFrom(Collider[] colliders)
    {
        var bounds = colliders[0].bounds;

        foreach (var colliderComponent in colliders)
        {
            bounds.Encapsulate(colliderComponent.bounds);
        }

        return bounds;
    }

    public static RaycastHit[] CharacterControllerCastAll(CharacterController characterController, Vector3 origin, Vector3 direction, float distance, int layerMask)
    {
        var scale = characterController.transform.lossyScale;
        var radius = characterController.radius * scale.x;
        var height = characterController.height * scale.y - (radius * 2);
        var center = characterController.center;
        center.Scale(scale);
        var point1 = origin + center + Vector3.down * (height / 2f);
        var point2 = point1 + Vector3.up * height;

        return Physics.CapsuleCastAll(point1, point2, radius, direction, distance, layerMask);
        //return Physics.RaycastAll(origin, direction, distance, layerMask);
    }

    public static RaycastHit[] CharacterControllerCastAll(CharacterController characterController, Vector3 origin, Vector3 direction, int layerMask)
    {
        var scale = characterController.transform.lossyScale;
        var radius = characterController.radius * scale.x;
        var height = characterController.height * scale.y - (radius * 2);
        var center = characterController.center;
        center.Scale(scale);
        var point1 = origin + center + Vector3.down * (height / 2f);
        var point2 = point1 + Vector3.up * height;

        return Physics.CapsuleCastAll(point1, point2, radius, direction, Mathf.Infinity, layerMask);
        //return Physics.RaycastAll(origin, direction, distance, layerMask);
    }

    public static bool CharacterControllerCast(CharacterController characterController, Vector3 origin, Vector3 direction, out RaycastHit hitInfo, int layerMask)
    {
        var scale = characterController.transform.lossyScale;
        var radius = characterController.radius * scale.x;
        var height = characterController.height * scale.y - (radius * 2);
        var center = characterController.center;
        center.Scale(scale);
        var point1 = origin + center + Vector3.down * (height / 2f);
        var point2 = point1 + Vector3.up * height;

        return Physics.CapsuleCast(point1, point2, radius, direction, out hitInfo, Mathf.Infinity, layerMask);
    }

    public static bool CharacterControllerCast(CharacterController characterController, Vector3 origin, Vector3 direction, float distance, int layerMask)
    {
        var scale = characterController.transform.lossyScale;
        var radius = characterController.radius * scale.x;
        var height = characterController.height * scale.y - (radius * 2);
        var center = characterController.center;
        center.Scale(scale);
        var point1 = origin + center + Vector3.down * (height / 2f);
        var point2 = point1 + Vector3.up * height;

        return Physics.CapsuleCast(point1, point2, radius, direction, distance, layerMask);
    }

    public static void PlayParticleSystem(ParticleSystem particleSystem, Vector3? position = null)
    {
        if (particleSystem == null)
            return;

        if (position != null)
            particleSystem.transform.position = position.Value;

        // Reset particle system interpolation
        // See: http://forum.unity3d.com/threads/134283-Moving-a-ParticleSystem-why-does-Unity-interpolate-its-position
        particleSystem.gameObject.SetActive(false);
        particleSystem.gameObject.SetActive(true);

        particleSystem.Play();
    }

    public static void StopParticleSystem(ParticleSystem particleSystem)
    {
        if (particleSystem != null)
            particleSystem.Stop();
    }

    public static void SpawnParticleSystem(ParticleSystem particleSystemPrefab, Vector3 position)
    {
        if (particleSystemPrefab == null)
            return;

        Object.Instantiate(particleSystemPrefab.gameObject, position, particleSystemPrefab.transform.rotation);
    }

    public static float RealDeltaTime { get { return (Time.timeScale > 0) ? (Time.deltaTime / Time.timeScale) : 0; } }
	public static float RealSmoothDeltaTime { get { return (Time.timeScale > 0) ? (Time.smoothDeltaTime / Time.timeScale) : 0; } }
}
